@@ -15,6 +15,24 @@
  *
  * You should have received a copy of the GNU General Public License along with
  * OpenPACE.  If not, see <http://www.gnu.org/licenses/>.
+ *
+ * Additional permission under GNU GPL version 3 section 7
+ *
+ * If you modify this Program, or any covered work, by linking or combining it
+ * with OpenSSL (or a modified version of that library), containing
+ * parts covered by the terms of OpenSSL's license, the licensors of
+ * this Program grant you additional permission to convey the resulting work.
+ * Corresponding Source for a non-source form of such a combination shall include
+ * the source code for the parts of OpenSSL used as well as that of the
+ * covered work.
+ *
+ * If you modify this Program, or any covered work, by linking or combining it
+ * with OpenSC (or a modified version of that library), containing
+ * parts covered by the terms of OpenSC's license, the licensors of
+ * this Program grant you additional permission to convey the resulting work. 
+ * Corresponding Source for a non-source form of such a combination shall include
+ * the source code for the parts of OpenSC used as well as that of the
+ * covered work.
  */
 
 /**
@@ -344,50 +362,111 @@ err:
     return out;
 }
 
+ASN1_OCTET_STRING *get_raw_authorizations(const struct gengetopt_args_info *cmdline)
+{
+    int ok = 0;
+    size_t i, hex_len, binary_len;
+    unsigned char *binary = NULL;
+    char *hex = cmdline->chat_arg;
+    ASN1_OCTET_STRING *out = NULL;
+
+    hex_len = strlen(hex);
+    if (hex_len % 2) {
+        err("hex string needs even number of bytes");
+    }
+    binary_len = hex_len / 2;
+
+    binary = calloc(sizeof *binary, binary_len);
+    if (!binary)
+        goto err;
+
+    for (i = 0; i < hex_len; i++) {
+        char c = *hex;
+        if (c == 0)
+            break;
+        if ((c >= '0') && (c <= '9'))
+            c -= '0';
+        else if ((c >= 'A') && (c <= 'F'))
+            c = c - 'A' + 10;
+        else if ((c >= 'a') && (c <= 'f'))
+            c = c - 'a' + 10;
+        else {
+            err("non-hex digit");
+        }
+        if (i & 1)
+            binary[i / 2] |= c;
+        else
+            binary[i / 2] = (c << 4);
+        hex++;
+    }
+
+    out = ASN1_OCTET_STRING_new();
+    if (!out || !ASN1_OCTET_STRING_set(out,
+                binary, binary_len))
+        goto err;
+
+err:
+    free(binary);
+
+    return out;
+}
+
 static CVC_CHAT *get_chat(const struct gengetopt_args_info *cmdline, CVC_CERT *signer)
 {
     CVC_CHAT *chat = NULL;
     int terminal_type = NID_undef;
+    size_t type_arg_len;
 
     if (!cmdline)
         goto err;
+    type_arg_len = strlen(cmdline->type_arg);
 
     chat = CVC_CHAT_new();
     if (!chat)
         goto err;
 
-    switch (cmdline->type_arg) {
-        case type_arg_at:
-            terminal_type = NID_id_AT;
-            break;
-        case type_arg_is:
+    if (strlen("at") == type_arg_len
+            && 0 == strcmp(cmdline->type_arg, "at")) {
+        terminal_type = NID_id_AT;
+        chat->terminal_type = OBJ_nid2obj(NID_id_AT);
+    } else if (strlen("is") == type_arg_len
+            && 0 == strcmp(cmdline->type_arg, "is")) {
             terminal_type = NID_id_IS;
-            break;
-        case type_arg_st:
+            chat->terminal_type = OBJ_nid2obj(NID_id_IS);
+    } else if (strlen("st") == type_arg_len
+            && 0 == strcmp(cmdline->type_arg, "st")) {
             terminal_type = NID_id_ST;
-            break;
-        case type_arg_derived_from_signer:
-            if (!signer || !signer->body || !signer->body->chat
-                    || !signer->body->chat->terminal_type)
-                err("type of signer is missing");
-            terminal_type = OBJ_obj2nid(signer->body->chat->terminal_type);
-            break;
-        default:
-            err("unhandled type of terminal");
+            chat->terminal_type = OBJ_nid2obj(NID_id_ST);
+    } else if (strlen("derived_from_signer") == type_arg_len
+            && 0 == strcmp(cmdline->type_arg, "derived_from_signer")) {
+        if (!signer || !signer->body || !signer->body->chat
+                || !signer->body->chat->terminal_type)
+            err("type of signer is missing");
+        terminal_type = OBJ_obj2nid(signer->body->chat->terminal_type);
+        chat->terminal_type = OBJ_dup(signer->body->chat->terminal_type);
+    } else {
+        terminal_type = OBJ_txt2nid(cmdline->type_arg);
+        chat->terminal_type = OBJ_txt2obj(cmdline->type_arg, 0);
     }
 
+    if (chat->relative_authorization)
+        ASN1_OCTET_STRING_free(chat->relative_authorization);
+
     if        (terminal_type == NID_id_AT) {
         chat->relative_authorization = get_at_authorizations(cmdline);
     } else if (terminal_type == NID_id_IS) {
-            chat->relative_authorization = get_is_authorizations(cmdline);
+        chat->relative_authorization = get_is_authorizations(cmdline);
     } else if (terminal_type == NID_id_ST) {
         chat->relative_authorization = get_st_authorizations(cmdline);
-    } else {
-        err("unhandled type of terminal");
     }
-    chat->terminal_type = OBJ_nid2obj(terminal_type);
-    if (!chat->terminal_type)
-        goto err;
+
+    if (cmdline->chat_arg) {
+        if (chat->relative_authorization) {
+            ASN1_OCTET_STRING_free(chat->relative_authorization);
+            chat->relative_authorization = get_raw_authorizations(cmdline);
+        }
+    }
+
 
 err:
     return chat;
@@ -476,7 +555,6 @@ CVC_CERTIFICATE_DESCRIPTION *create_certificate_description(const struct gengeto
     unsigned char *desc_data = NULL;
     size_t len = 0, ext_len = 0, desc_data_len = 0;
     int desc_type = NID_undef;
-    void *asn1 = NULL;
 
     if (cmdline->cert_desc_arg) {
         len = strlen(cmdline->cert_desc_arg);
@@ -507,39 +585,14 @@ CVC_CERTIFICATE_DESCRIPTION *create_certificate_description(const struct gengeto
         if (0 != read_file(cmdline->cert_desc_arg, &desc_data, &desc_data_len)) {
             goto err;
         }
-        if        (desc_type == NID_id_plainFormat) {
-                asn1 = ASN1_UTF8STRING_new();
-        } else if (desc_type == NID_id_htmlFormat) {
-            asn1 = ASN1_IA5STRING_new();
-        } else {
-            /* desc_type == NID_id_pdfFormat */
-            asn1 = ASN1_OCTET_STRING_new();
-        }
-        if (!asn1 || !ASN1_OCTET_STRING_set(asn1, desc_data, desc_data_len))
+        desc->termsOfUsage = ASN1_OCTET_STRING_new();
+        if (!desc->termsOfUsage || !ASN1_OCTET_STRING_set(
+                    desc->termsOfUsage, desc_data, desc_data_len))
             goto err;
-#ifdef HAVE_PATCHED_OPENSSL
-        switch (desc_type) {
-            case NID_id_plainFormat:
-                desc->termsOfUsage.plainTerms = asn1;
-                break;
-            case NID_id_htmlFormat:
-                desc->termsOfUsage.htmlTerms = asn1;
-                break;
-            case NID_id_pdfFormat:
-                desc->termsOfUsage.pdfTerms = asn1;
-                break;
-            default:
-                goto err;
-        }
-#else
-        desc->termsOfUsage.other = ASN1_TYPE_new();
-        if (!desc->termsOfUsage.other)
-            goto err;
-        ASN1_TYPE_set(desc->termsOfUsage.other, V_ASN1_SEQUENCE, asn1);
-#endif
 
         if (cmdline->issuer_name_arg) {
-            desc->issuerName = ASN1_UTF8STRING_new();
+            if (!desc->issuerName)
+                desc->issuerName = ASN1_UTF8STRING_new();
             if (!desc->issuerName
                     || !ASN1_OCTET_STRING_set(desc->issuerName,
                          (const unsigned char *) cmdline->issuer_name_arg,
@@ -589,7 +642,6 @@ int main(int argc, char *argv[])
 {
     CVC_CERT *cert = NULL;
     CVC_CERT *sign_as_cert = NULL;
-    CVC_CERT_BODY *body = NULL;
     CVC_CERTIFICATE_DESCRIPTION *desc = NULL;
     CVC_CERT_REQUEST *request = NULL;
     CVC_CERT_AUTHENTICATION_REQUEST *authentication = NULL;
@@ -600,7 +652,7 @@ int main(int argc, char *argv[])
     unsigned char *body_p = NULL, *cert_buf = NULL, *term_key_buf = NULL, *desc_buf = NULL;
     size_t car_len = 0, cert_len = 0;
     time_t loc;
-    const struct tm *utc_tm;
+    struct tm utc_tm;
     char string[80];
     const char *out = NULL;
     char basename[70];
@@ -615,11 +667,14 @@ int main(int argc, char *argv[])
         return 1;
     }
 
-    body = CVC_CERT_BODY_new();
     cert = CVC_CERT_new();
-    if (!cert || !body)
+    if (!cert)
         goto err;
-    cert->body = body;
+    if (!cert->body) {
+        cert->body = CVC_CERT_BODY_new();
+        if (!cert->body)
+            goto err;
+    }
 
 
     /* read certificate signing request */
@@ -638,9 +693,10 @@ int main(int argc, char *argv[])
 
 
     /* write profile identifier fixed to 0 ("version 1") */
-    body->certificate_profile_identifier = ASN1_INTEGER_new();
-    if (!body->certificate_profile_identifier
-            || !ASN1_INTEGER_set(body->certificate_profile_identifier, 0))
+    if (!cert->body->certificate_profile_identifier)
+        cert->body->certificate_profile_identifier = ASN1_INTEGER_new();
+    if (!cert->body->certificate_profile_identifier
+            || !ASN1_INTEGER_set(cert->body->certificate_profile_identifier, 0))
         goto err;
 
 
@@ -664,29 +720,31 @@ int main(int argc, char *argv[])
                 car_len = request->body->certificate_holder_reference->length;
             }
         }
-        body->certificate_authority_reference = ASN1_UTF8STRING_new();
-        if (!body->certificate_authority_reference
-                || !ASN1_OCTET_STRING_set(body->certificate_authority_reference, car, car_len))
+        if (!cert->body->certificate_authority_reference)
+            cert->body->certificate_authority_reference = ASN1_UTF8STRING_new();
+        if (!cert->body->certificate_authority_reference
+                || !ASN1_OCTET_STRING_set(cert->body->certificate_authority_reference, car, car_len))
             goto err;
     } else {
-        body->certificate_authority_reference = (ASN1_UTF8STRING *) ASN1_STRING_dup((ASN1_STRING *) request->body->certificate_authority_reference);
-        if (!body->certificate_authority_reference)
+        cert->body->certificate_authority_reference = (ASN1_UTF8STRING *) ASN1_STRING_dup((ASN1_STRING *) request->body->certificate_authority_reference);
+        if (!cert->body->certificate_authority_reference)
             goto err;
     }
 
 
     /* write CHR */
     if (cmdline.manual_mode_counter) {
-        body->certificate_holder_reference = ASN1_UTF8STRING_new();
-        if (!body->certificate_holder_reference
-                || !ASN1_OCTET_STRING_set(body->certificate_holder_reference,
+        if (!cert->body->certificate_holder_reference)
+            cert->body->certificate_holder_reference = ASN1_UTF8STRING_new();
+        if (!cert->body->certificate_holder_reference
+                || !ASN1_OCTET_STRING_set(cert->body->certificate_holder_reference,
                     (unsigned char *) cmdline.chr_arg, strlen(cmdline.chr_arg)))
             goto err;
-        strncpy(basename, cmdline.chr_arg, sizeof basename);
+        strncpy(basename, cmdline.chr_arg, (sizeof basename) - 1);
         basename[sizeof basename - 1] = '\0';
     } else {
-        body->certificate_holder_reference = (ASN1_UTF8STRING *) ASN1_STRING_dup((ASN1_STRING *) request->body->certificate_holder_reference);
-        if (!body->certificate_holder_reference)
+        cert->body->certificate_holder_reference = (ASN1_UTF8STRING *) ASN1_STRING_dup((ASN1_STRING *) request->body->certificate_holder_reference);
+        if (!cert->body->certificate_holder_reference)
             goto err;
         memcpy(basename, (char *) request->body->certificate_holder_reference->data,
                 sizeof basename < request->body->certificate_holder_reference->length ?
@@ -747,52 +805,102 @@ int main(int argc, char *argv[])
 
 
         /* write public key */
-        body->public_key = get_cvc_pubkey(&cmdline, term_key);
+        if (cert->body->public_key)
+            CVC_PUBKEY_free(cert->body->public_key);
+        cert->body->public_key = get_cvc_pubkey(&cmdline, term_key);
     } else {
         /* write public key */
-        body->public_key = CVC_PUBKEY_dup(request->body->public_key);
+        if (cert->body->public_key)
+            CVC_PUBKEY_free(cert->body->public_key);
+        if (cmdline.role_arg == role_arg_cvca) {
+            cert->body->public_key = CVC_PUBKEY_dup(request->body->public_key);
+        } else {
+            int nid = OBJ_obj2nid(request->body->public_key->oid);
+            if (       nid == NID_id_TA_ECDSA_SHA_1
+                    || nid == NID_id_TA_ECDSA_SHA_224
+                    || nid == NID_id_TA_ECDSA_SHA_256
+                    || nid == NID_id_TA_ECDSA_SHA_384
+                    || nid == NID_id_TA_ECDSA_SHA_512) {
+                cert->body->public_key = CVC_PUBKEY_new();
+                if (!cert->body->public_key)
+                    goto err;
+                /* copy only the public key without explicit parameters */
+                cert->body->public_key->oid = OBJ_dup(request->body->public_key->oid);
+                cert->body->public_key->cont6 = ASN1_OCTET_STRING_dup(request->body->public_key->cont6);
+                if (!cert->body->public_key->oid
+                        || !cert->body->public_key->cont6)
+                    goto err;
+            } else if (nid == NID_id_TA_RSA_v1_5_SHA_1
+                    || nid == NID_id_TA_RSA_v1_5_SHA_256
+                    || nid == NID_id_TA_RSA_v1_5_SHA_512
+                    || nid == NID_id_TA_RSA_PSS_SHA_1
+                    || nid == NID_id_TA_RSA_PSS_SHA_256
+                    || nid == NID_id_TA_RSA_PSS_SHA_512) {
+                /* copy only the public key without explicit parameters */
+                cert->body->public_key->oid = OBJ_dup(request->body->public_key->oid);
+                cert->body->public_key->cont1 = ASN1_OCTET_STRING_dup(request->body->public_key->cont1);
+                cert->body->public_key->cont2 = ASN1_OCTET_STRING_dup(request->body->public_key->cont2);
+                if (!cert->body->public_key->oid
+                        || !cert->body->public_key->cont1
+                        || !cert->body->public_key->cont2)
+                    goto err;
+            } else {
+                /* unknown mechanism, just copy everything */
+                cert->body->public_key = CVC_PUBKEY_dup(request->body->public_key);
+            }
+        }
     }
-    if (!body->public_key)
+    if (!cert->body->public_key)
         goto err;
 
 
     /* write effective date */
     if (!cmdline.issued_given) {
         time(&loc);
-        utc_tm = gmtime(&loc);
-        if (!utc_tm || utc_tm->tm_year < 100 || utc_tm->tm_year > 2000
-                || utc_tm->tm_mon < 0 || utc_tm->tm_mon > 11
-                || utc_tm->tm_mday < 0 || utc_tm->tm_mday > 31)
+#ifdef _WIN32
+		if (0 != gmtime_s(&utc_tm, &loc))
+			goto err;
+#else
+		if (NULL == gmtime_r(&loc, &utc_tm))
+			goto err;
+#endif
+        if (utc_tm.tm_year < 100 || utc_tm.tm_year > 2000
+                || utc_tm.tm_mon < 0 || utc_tm.tm_mon > 11
+                || utc_tm.tm_mday < 0 || utc_tm.tm_mday > 31)
             goto err;
-        string[0] = (char) (utc_tm->tm_year-100)/10;
-        string[1] = (char) utc_tm->tm_year%10;
-        string[2] = (char) (utc_tm->tm_mon+1)/10;
-        string[3] = (char) (utc_tm->tm_mon+1)%10;
-        string[4] = (char) utc_tm->tm_mday/10;
-        string[5] = (char) utc_tm->tm_mday%10;
+        string[0] = (char) (utc_tm.tm_year-100)/10;
+        string[1] = (char) utc_tm.tm_year%10;
+        string[2] = (char) (utc_tm.tm_mon+1)/10;
+        string[3] = (char) (utc_tm.tm_mon+1)%10;
+        string[4] = (char) utc_tm.tm_mday/10;
+        string[5] = (char) utc_tm.tm_mday%10;
     } else {
         if (!to_bcd(cmdline.issued_arg, (unsigned char *) string, 6))
             goto err;
     }
-    body->certificate_effective_date = ASN1_OCTET_STRING_new();
-    if (!body->certificate_effective_date
-            || !ASN1_OCTET_STRING_set(body->certificate_effective_date,
+    if (!cert->body->certificate_effective_date)
+        cert->body->certificate_effective_date = ASN1_OCTET_STRING_new();
+    if (!cert->body->certificate_effective_date
+            || !ASN1_OCTET_STRING_set(cert->body->certificate_effective_date,
                 (unsigned char *) string, 6))
         goto err;
 
 
     /* write expiration date */
-    body->certificate_expiration_date = ASN1_OCTET_STRING_new();
-    if (!body->certificate_expiration_date
+    if (!cert->body->certificate_expiration_date)
+        cert->body->certificate_expiration_date = ASN1_OCTET_STRING_new();
+    if (!cert->body->certificate_expiration_date
             || !to_bcd(cmdline.expires_arg, (unsigned char *) string, 6)
-            || !ASN1_OCTET_STRING_set(body->certificate_expiration_date,
+            || !ASN1_OCTET_STRING_set(cert->body->certificate_expiration_date,
                 (unsigned char *) string, 6))
         goto err;
 
 
     /* write chat */
-    body->chat = get_chat(&cmdline, sign_as_cert);
-    if (!body->chat)
+    if (cert->body->chat)
+        CVC_CHAT_free(cert->body->chat);
+    cert->body->chat = get_chat(&cmdline, sign_as_cert);
+    if (!cert->body->chat)
         goto err;
 
 
@@ -829,7 +937,7 @@ int main(int argc, char *argv[])
 
 
     /* sign body */
-    body_len = i2d_CVC_CERT_BODY(body, &body_p);
+    body_len = i2d_CVC_CERT_BODY(cert->body, &body_p);
     if (body_len <= 0)
         goto err;
     body_buf = BUF_MEM_create_init(body_p, (size_t) body_len);
@@ -847,7 +955,8 @@ int main(int argc, char *argv[])
 
 
     /* assamble everything */
-    cert->signature = ASN1_OCTET_STRING_new();
+    if (!cert->signature)
+        cert->signature = ASN1_OCTET_STRING_new();
     if (!cert->signature
             || !ASN1_OCTET_STRING_set(cert->signature,
                 (unsigned char *) signature->data, signature->length))
@@ -875,9 +984,6 @@ err:
     cmdline_parser_free (&cmdline);
     if (cert) {
         CVC_CERT_free(cert);
-    } else {
-        if (body)
-            CVC_CERT_BODY_free(body);
     }
     if (sign_as_cert)
         CVC_CERT_free(sign_as_cert);
