@@ -15,6 +15,24 @@
  *
  * You should have received a copy of the GNU General Public License along with
  * OpenPACE.  If not, see <http://www.gnu.org/licenses/>.
+ *
+ * Additional permission under GNU GPL version 3 section 7
+ *
+ * If you modify this Program, or any covered work, by linking or combining it
+ * with OpenSSL (or a modified version of that library), containing
+ * parts covered by the terms of OpenSSL's license, the licensors of
+ * this Program grant you additional permission to convey the resulting work.
+ * Corresponding Source for a non-source form of such a combination shall include
+ * the source code for the parts of OpenSSL used as well as that of the
+ * covered work.
+ *
+ * If you modify this Program, or any covered work, by linking or combining it
+ * with OpenSC (or a modified version of that library), containing
+ * parts covered by the terms of OpenSC's license, the licensors of
+ * this Program grant you additional permission to convey the resulting work. 
+ * Corresponding Source for a non-source form of such a combination shall include
+ * the source code for the parts of OpenSC used as well as that of the
+ * covered work.
  */
 
 /**
@@ -60,10 +78,10 @@ struct eac_ctx;
 enum cvc_terminal_role {
     /** @brief Terminal (inspection system/authentication terminal/signature terminal) */
     CVC_Terminal = 0,
-    /** @brief Document Verifier (official domestic) */
-    CVC_DV,
     /** @brief Document Verifier (non-official/foreign) */
     CVC_DocVer, /* XXX Ugly */
+    /** @brief Document Verifier (official domestic) */
+    CVC_DV,
     /** @brief Country Verifying Certificate Authority */
     CVC_CVCA
 };
@@ -227,22 +245,7 @@ typedef struct cvc_commcert_seq_st {
     ASN1_PRINTABLESTRING *subjectURL;
      /** @brief Terms of Usage of the Service holding the certificate. May be
      *  formatted as either plain text, HTML or PDF */
-    union {
-#ifdef HAVE_PATCHED_OPENSSL
-        /** @brief Plain text Terms of Usage */
-        ASN1_UTF8STRING *plainTerms;
-        /** @brief HTML formatted Terms of Usage */
-        ASN1_IA5STRING *htmlTerms;
-        /** @brief PDF formatted Terms of Usage */
-        ASN1_OCTET_STRING *pdfTerms;
-#else
-        /** @brief just a placeholder */
-        ASN1_OCTET_STRING *unused;
-#endif
-        /** @brief Otherwise formatted Terms of Usage (not specified) */
-        ASN1_TYPE *other;
-    } termsOfUsage;
-
+    ASN1_OCTET_STRING *termsOfUsage;
     /** @brief Not used */
     ASN1_PRINTABLESTRING *redirectURL;
     /** @brief Contains hash values of admissible X.509 certificates of the remote
