@@ -15,6 +15,24 @@
  *
  * You should have received a copy of the GNU General Public License along with
  * OpenPACE.  If not, see <http://www.gnu.org/licenses/>.
+ *
+ * Additional permission under GNU GPL version 3 section 7
+ *
+ * If you modify this Program, or any covered work, by linking or combining it
+ * with OpenSSL (or a modified version of that library), containing
+ * parts covered by the terms of OpenSSL's license, the licensors of
+ * this Program grant you additional permission to convey the resulting work.
+ * Corresponding Source for a non-source form of such a combination shall include
+ * the source code for the parts of OpenSSL used as well as that of the
+ * covered work.
+ *
+ * If you modify this Program, or any covered work, by linking or combining it
+ * with OpenSC (or a modified version of that library), containing
+ * parts covered by the terms of OpenSC's license, the licensors of
+ * this Program grant you additional permission to convey the resulting work. 
+ * Corresponding Source for a non-source form of such a combination shall include
+ * the source code for the parts of OpenSC used as well as that of the
+ * covered work.
  */
 
 /**
@@ -153,6 +171,8 @@ cipher(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *type, ENGINE *impl,
             goto err;
     }
 
+    if (!EVP_CIPHER_CTX_cipher(tmp_ctx))
+        goto err;
     flags = EVP_CIPHER_flags(EVP_CIPHER_CTX_cipher(tmp_ctx));
 
     if (flags & EVP_CIPH_NO_PADDING) {
@@ -462,11 +482,13 @@ int update_iv(KA_CTX *ctx, EVP_CIPHER_CTX *cipher_ctx, const BIGNUM *ssc)
             ctx->cipher = oldcipher;
             if (!ivbuf)
                 goto err;
-            p = OPENSSL_realloc(ctx->iv, ivbuf->length);
-            if (!p)
-                goto err;
-            ctx->iv = p;
-            memcpy(ctx->iv, ivbuf->data, ivbuf->length);
+            if (0 != ivbuf->length) {
+                p = OPENSSL_realloc(ctx->iv, ivbuf->length);
+                if (!p)
+                    goto err;
+                ctx->iv = p;
+                memcpy(ctx->iv, ivbuf->data, ivbuf->length);
+            }
             break;
 
         case NID_des_ede_cbc:
@@ -596,6 +618,7 @@ EVP_PKEY_dup(EVP_PKEY *key)
 
     switch (EVP_PKEY_base_id(key)) {
         case EVP_PKEY_DH:
+        case EVP_PKEY_DHX:
             dh_in = EVP_PKEY_get1_DH(key);
             if (!dh_in)
                 goto err;
@@ -651,6 +674,8 @@ err:
         EC_KEY_free(ec_in);
     if (rsa_in)
         RSA_free(rsa_in);
+    if (out)
+        EVP_PKEY_free(out);
 
     return NULL;
 }
@@ -768,8 +793,10 @@ Comp(EVP_PKEY *key, const BUF_MEM *pub, BN_CTX *bn_ctx, EVP_MD_CTX *md_ctx)
 
     check((key && pub), "Invalid arguments");
 
+    BN_CTX_start(bn_ctx);
     switch (EVP_PKEY_base_id(key)) {
         case EVP_PKEY_DH:
+        case EVP_PKEY_DHX:
             out = hash(EVP_sha1(), md_ctx, NULL, pub);
             break;
 
@@ -786,7 +813,7 @@ Comp(EVP_PKEY *key, const BUF_MEM *pub, BN_CTX *bn_ctx, EVP_MD_CTX *md_ctx)
             if(!ecp || !x || !y
                     || !EC_POINT_oct2point(group, ecp,
                         (unsigned char *) pub->data, pub->length, bn_ctx)
-                    || !EC_POINT_get_affine_coordinates_GF2m(group, ecp, x, y, bn_ctx))
+                    || !EC_POINT_get_affine_coordinates(group, ecp, x, y, bn_ctx))
                 goto err;
 
             out = BUF_MEM_create(BN_num_bytes(x));
@@ -806,10 +833,7 @@ err:
      * structure */
     if (ec)
         EC_KEY_free(ec);
-    if (x)
-        BN_free(x);
-    if (y)
-        BN_free(y);
+    BN_CTX_end(bn_ctx);
 
     return out;
 }
@@ -912,6 +936,7 @@ EVP_PKEY_set_keys(EVP_PKEY *evp_pkey,
             break;
 
         case EVP_PKEY_DH:
+        case EVP_PKEY_DHX:
             dh = EVP_PKEY_get1_DH(evp_pkey);
             if (!dh)
                 goto err;
@@ -932,7 +957,7 @@ EVP_PKEY_set_keys(EVP_PKEY *evp_pkey,
             break;
 
         default:
-            log_err("Unknown type of key");
+            log_err("Unknown type of key %d", EVP_PKEY_base_id(evp_pkey));
             goto err;
             break;
     }
@@ -965,6 +990,7 @@ get_pubkey(EVP_PKEY *key, BN_CTX *bn_ctx)
 
     switch (EVP_PKEY_base_id(key)) {
         case EVP_PKEY_DH:
+        case EVP_PKEY_DHX:
             dh = EVP_PKEY_get1_DH(key);
             check_return(dh, "no DH key");
 
